@@ -1,24 +1,18 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
-###
-### Author:  Chris Iatrou (ichrispa@core-vector.net)
-### Version: rev 13
-###
-### This program was created for educational purposes and has been
-### contributed to the open62541 project by the author. All licensing
-### terms for this source is inherited by the terms and conditions
-### specified for by the open62541 project (see the projects readme
-### file for more information on the LGPL terms and restrictions).
-###
-### This program is not meant to be used in a production environment. The
-### author is not liable for any complications arising due to the use of
-### this program.
-###
+### This Source Code Form is subject to the terms of the Mozilla Public
+### License, v. 2.0. If a copy of the MPL was not distributed with this
+### file, You can obtain one at http://mozilla.org/MPL/2.0/.
+
+###    Copyright 2014-2015 (c) TU-Dresden (Author: Chris Iatrou)
+###    Copyright 2014-2017 (c) Fraunhofer IOSB (Author: Julius Pfrommer)
+###    Copyright 2016-2017 (c) Stefan Profanter, fortiss GmbH
+
 
 import sys
 import logging
-from datatypes import *
+from datatypes import QualifiedName, LocalizedText, NodeId, String, Value, valueIsInternalType
 
 __all__ = ['Reference', 'RefOrAlias', 'Node', 'ReferenceTypeNode',
            'ObjectNode', 'VariableNode', 'VariableTypeNode',
@@ -80,6 +74,7 @@ class Node(object):
         self.modelUri = None
         self.parent = None
         self.parentReference = None
+        self.namespaceMapping = {}
 
     def __str__(self):
         return self.__class__.__name__ + "(" + str(self.id) + ")"
@@ -190,6 +185,7 @@ class Node(object):
             ref.referenceType.ns = nsMapping[ref.referenceType.ns]
             new_refs.add(ref)
         self.references = new_refs
+        self.namespaceMapping = nsMapping
 
 class ReferenceTypeNode(Node):
     def __init__(self, xmlelement=None):
@@ -259,6 +255,8 @@ class VariableNode(Node):
                 self.dataType = RefOrAlias(av)
             elif  at == "ArrayDimensions":
                 self.arrayDimensions = av.split(",")
+            elif at == "Historizing":
+                self.historizing = "false" not in av.lower()
 
         for x in xmlelement.childNodes:
             if x.nodeType != x.ELEMENT_NODE:
@@ -288,13 +286,12 @@ class VariableNode(Node):
         if dataTypeNode is None:
             return False
 
-        # FIXME: Don't build at all or allocate "defaults"? I'm for not building at all.
         if self.xmlValueDef is None:
             #logger.warn("Variable " + self.browseName() + "/" + str(self.id()) + " is not initialized. No memory will be allocated.")
             return False
 
         self.value = Value()
-        self.value.parseXMLEncoding(self.xmlValueDef, dataTypeNode, self, nodeset.namespaceMapping[self.modelUri])
+        self.value.parseXMLEncoding(self.xmlValueDef, dataTypeNode, self, nodeset.parser)
         return True
 
 
@@ -358,7 +355,7 @@ class DataTypeNode(Node):
         1) A DataType can be a structure of fields, each field having a name and a type.
            The type must be either an encodable builtin node (ex. UInt32) or point to
            another DataType node that inherits its encoding from a builtin type using
-           a inverse "hasSubtype" (hasSuperType) reference.
+           an inverse "hasSubtype" (hasSuperType) reference.
         2) A DataType may be an enumeration, in which each field has a name and a numeric
            value.
         The definition is stored as an ordered list of tuples. Depending on which
@@ -426,7 +423,7 @@ class DataTypeNode(Node):
             return self.__baseTypeEncoding__
 
 
-    def buildEncoding(self, nodeset, indent=0, force=False):
+    def buildEncoding(self, nodeset, indent=0, force=False, namespaceMapping=None):
         """ buildEncoding() determines the structure and aliases used for variables
             of this DataType.
 
@@ -504,14 +501,15 @@ class DataTypeNode(Node):
         if self.__xmlDefinition__ is None:
             if parentType is not None:
                 logger.debug( prefix + "Attempting definition using supertype " + str(targetNode.browseName) + " for DataType " + " " + str(self.browseName))
-                subenc = targetNode.buildEncoding(nodeset=nodeset, indent=indent+1)
+                subenc = targetNode.buildEncoding(nodeset=nodeset, indent=indent+1,
+                                                  namespaceMapping=namespaceMapping)
                 if not targetNode.isEncodable():
-                    self.__encodable__ = False
+                    self.__encodable__ = True
                 else:
-                    self.__baseTypeEncoding__ = self.__baseTypeEncoding__ + [self.browseName.name, subenc, None]
+                    self.__baseTypeEncoding__ = self.__baseTypeEncoding__ + [self.browseName.name, subenc, None , 'false']
             if len(self.__baseTypeEncoding__) == 0:
                 logger.debug(prefix + "No viable definition for " + str(self.browseName) + " " + str(self.id) + " found.")
-                self.__encodable__ = False
+                self.__encodable__ = True
 
             if indent==0:
                 if not self.__encodable__:
@@ -526,7 +524,7 @@ class DataTypeNode(Node):
         # An option set is at the same time also an enum, at least for the encoding below
         isOptionSet = parentType is not None and parentType.id.ns == 0 and parentType.id.i==12755
 
-        # We need to store the definition as ordered data, but can't use orderedDict
+        # We need to store the definition as ordered data, but cannot use orderedDict
         # for backward compatibility with Python 2.6 and 3.4
         enumDict = []
         typeDict = []
@@ -539,6 +537,8 @@ class DataTypeNode(Node):
                 enumVal = ""
                 valueRank = None
                 #symbolicName = None
+                arrayDimensions = None
+                isOptional = ""
                 for at,av in x.attributes.items():
                     if at == "DataType":
                         fdtype = str(av)
@@ -555,6 +555,13 @@ class DataTypeNode(Node):
                         isEnum = True
                     elif at == "ValueRank":
                         valueRank = int(av)
+                    elif at == "IsOptional":
+                        isOptional = str(av)
+                    elif at == "ArrayDimensions":
+                        arrayDimensions = int(av)
+                    elif at == "AllowSubTypes":
+                        # ignore
+                        continue
                     else:
                         logger.warn("Unknown Field Attribute " + str(at))
                 # This can either be an enumeration OR a structure, not both.
@@ -571,7 +578,8 @@ class DataTypeNode(Node):
                     # This might be a subtype... follow the node defined as datatype to find out
                     # what encoding to use
                     fdTypeNodeId = NodeId(fdtype)
-                    fdTypeNodeId.ns = nodeset.namespaceMapping[self.modelUri][fdTypeNodeId.ns]
+                    if namespaceMapping != None:
+                        fdTypeNodeId.ns = namespaceMapping[fdTypeNodeId.ns]
                     if not fdTypeNodeId in nodeset.nodes:
                         raise Exception("Node {} not found in nodeset".format(fdTypeNodeId))
                     dtnode = nodeset.nodes[fdTypeNodeId]
@@ -580,12 +588,16 @@ class DataTypeNode(Node):
                     typeDict.append([fname, dtnode])
                     fdtype = str(dtnode.browseName.name)
                     logger.debug( prefix + fname + " : " + fdtype + " -> " + str(dtnode.id))
-                    subenc = dtnode.buildEncoding(nodeset=nodeset, indent=indent+1)
-                    self.__baseTypeEncoding__ = self.__baseTypeEncoding__ + [[fname, subenc, valueRank]]
+                    subenc = dtnode.buildEncoding(nodeset=nodeset, indent=indent+1,
+                                                  namespaceMapping=namespaceMapping)
+                    if isOptional:
+                        self.__baseTypeEncoding__ = self.__baseTypeEncoding__ + [[fname, subenc, valueRank, 'true']]
+                    else:
+                        self.__baseTypeEncoding__ = self.__baseTypeEncoding__ + [[fname, subenc, valueRank, 'false']]
                     if not dtnode.isEncodable():
                         # If we inherit an encoding from an unencodable node, this node is
                         # also not encodable
-                        self.__encodable__ = False
+                        self.__encodable__ = True
                         break
 
         # If we used inheritance to determine an encoding without alias, there is a
@@ -597,9 +609,9 @@ class DataTypeNode(Node):
 
         if isOptionSet == True:
             self.__isOptionSet__ = True
-            subenc = parentType.buildEncoding(nodeset=nodeset)
+            subenc = parentType.buildEncoding(nodeset=nodeset, namespaceMapping=namespaceMapping)
             if not parentType.isEncodable():
-                self.__encodable__ = False
+                self.__encodable__ = True
             else:
                 self.__baseTypeEncoding__ = self.__baseTypeEncoding__ + [self.browseName.name, subenc, None]
                 self.__definition__ = enumDict
@@ -618,8 +630,8 @@ class DataTypeNode(Node):
             else:
                 logger.debug( "Encodable as: " + str(self.__baseTypeEncoding__))
                 self.__isEnum__ = False
-                self.__definition__ = typeDict
             logger.debug( "")
+        self.__definition__ = typeDict
         return self.__baseTypeEncoding__
 
 class ViewNode(Node):
